"""
This package contains the public API for eFLIPS-Depot.

It is to be used in conjunction with the
`eflips.model <https://github.com/mpm-tu-berlin/eflips-model>`_ package, where the Scenario is defined.

Notes on the usage of the API
-----------------------------

The following steps are recommended for using the API:

1. Check if there are already "driving events" in the database. They come from a "consumption simulation" and are
   associated with a vehicle. If there are no driving events, you may use the :func:`simple_consumption_simulation`
   (with ``initialize_vehicles=True``) to create them. This function will also initialize the vehicles in the database
   with the correct vehicle type and assign them to rotations.
2. Check if there is already a depot layout in the database. If there is not, you may use the
   :func:`generate_depot_layout` function to create a simple depot layout and plan.
3. Either use the :func:`simulate_scenario` function to run the whole simulation in one go, or use the following steps:
    a. Use the :func:`init_simulation` function to create a simulation host, which is a "black box" object containing
       all input data for the simulation.
    b. Use the :func:`run_simulation` function to run the simulation and obtain the results.
    c. Use the :func:`add_evaluation_to_database` function to add the results to the database.
4. For the results to be valid, the consumption simulation should now be run again.
    a. If you are using an external consumption model, run it again making sure it does not create new vehicles.
    b. Run the :func:`simple_consumption_simulation` function again, this time with ``initialize_vehicles=False``.
"""
import copy
import datetime
import itertools
import logging
import os
import warnings
from collections import OrderedDict
from datetime import timedelta
from enum import Enum
from math import ceil
from typing import Any, Dict, Optional, Union, List

import numpy as np
import sqlalchemy.orm
from eflips.model import (
    Area,
    Depot,
    Event,
    EventType,
    Rotation,
    Scenario,
    Trip,
    Vehicle,
)
from sqlalchemy.orm import Session
from sqlalchemy.sql import select

import eflips.depot
from eflips.depot import DepotEvaluation, ProcessStatus, SimulationHost, SimpleVehicle
from eflips.depot.api.private.depot import (
    create_simple_depot,
    delete_depots,
    depot_to_template,
    group_rotations_by_start_end_stop,
)
from eflips.depot.api.private.smart_charging import optimize_charging_events_even
from eflips.depot.api.private.util import (
    create_session,
    repeat_vehicle_schedules,
    start_and_end_times,
    vehicle_type_to_global_constants_dict,
    VehicleSchedule,
    check_depot_validity,
)


class SmartChargingStrategy(Enum):
    """Enum class for different smart charging strategies."""

    NONE = 0
    """
    Do not use smart charging.

    Buses are charged with the maximum power available, from the time they arrive at the depot
    until they are full (or leave the depot).
    """
    EVEN = 1
    """
    Use smart charging with an even distribution of charging power over the time the bus is at the depot.

    This aims to
    minimize the peak power demand.
    """
    MIN_PRICE = 2
    """
    Use smart charging in order to minimize the cost of charging.

    The price profile can be specified using the
    PRICE_PROFILE environment variable. If this is not set, the price is loaded using an API.
    """


def simple_consumption_simulation(
    scenario: Union[Scenario, int, Any],
    initialize_vehicles: bool,
    database_url: Optional[str] = None,
    calculate_timeseries: bool = False,
    terminus_deadtime: timedelta = timedelta(minutes=1),
) -> None:
    """
    A simple consumption simulation and vehicle initialization.

    Energy consumotion is calculated by multiplying the vehicle's total distance by a constant
    ``VehicleType.consumption``.

    If run with ``initialize_vehicles=True``, the method will also initialize the vehicles in the database with the
    correct vehicle type and assign them to rotations. If this is false, it will assume that there are already vehicle
    entries and ``Rotation.vehicle_id`` is already set.

    :param scenario: Either a :class:`eflips.model.Scenario` object containing the input data for the simulation. Or
        an integer specifying the ID of a scenario in the database. Or any other object that has an attribute
        ``id`` that is an integer. If no :class:`eflips.model.Scenario` object is passed, the ``database_url``
        parameter must be set to a valid database URL ot the environment variable ``DATABASE_URL`` must be set to a
        valid database URL.

    :param initialize_vehicles: A boolean flag indicating whether the vehicles should be initialized in the database.
        When running this function for the first time, this should be set to True. When running this function again
        after the vehicles have been initialized, this should be set to False.

    :param database_url: An optional database URL. If no database URL is passed and the `scenario` parameter is not a
        :class:`eflips.model.Scenario` object, the environment variable `DATABASE_URL` must be set to a
        valid database URL.

    :param calculate_timeseries: A boolean flag indicating whether the timeseries should be calculated. If this is set
        to True, the SoC at each stop is calculated and added to the "timeseries" column of the Event table. If this
        is set to False, the "timeseries" column of the Event table will be set to ``None``. Setting this to false
        may significantly speed up the simulation.

    :param terminus_deadtime: The total deadtime taken to both attach and detach the charging cable at the terminus.
                              If the total deadtime is greater than the time between the arrival and departure of the
                              vehicle at the terminus, the vehicle will not be able to charge at the terminus.

    :return: Nothing. The results are added to the database.
    """
    logger = logging.getLogger(__name__)

    with create_session(scenario, database_url) as (session, scenario):
        rotations = (
            session.query(Rotation)
            .filter(Rotation.scenario_id == scenario.id)
            .order_by(Rotation.id)
            .options(sqlalchemy.orm.joinedload(Rotation.trips).joinedload(Trip.route))
            .options(sqlalchemy.orm.joinedload(Rotation.vehicle_type))
            .options(sqlalchemy.orm.joinedload(Rotation.vehicle))
        )
        if initialize_vehicles:
            for rotation in rotations:
                vehicle = Vehicle(
                    vehicle_type_id=rotation.vehicle_type_id,
                    scenario_id=scenario.id,
                    name=f"Vehicle for rotation {rotation.id}",
                )
                session.add(vehicle)
                rotation.vehicle = vehicle

                # Additionally, add a short STANDBY event with 100% SoC immediately before the first trip
                first_trip_start = rotation.trips[0].departure_time
                standby_start = first_trip_start - timedelta(seconds=1)
                standby_event = Event(
                    scenario_id=scenario.id,
                    vehicle_type_id=rotation.vehicle_type_id,
                    vehicle=vehicle,
                    station_id=rotation.trips[0].route.departure_station_id,
                    subloc_no=0,
                    time_start=standby_start,
                    time_end=first_trip_start,
                    soc_start=1,
                    soc_end=1,
                    event_type=EventType.CHARGING_OPPORTUNITY,
                    description=f"DUMMY Initial standby event for rotation {rotation.id}.",
                    timeseries=None,
                )
                session.add(standby_event)
        else:
            for rotation in rotations:
                if rotation.vehicle is None:
                    raise ValueError(
                        "The rotation does not have a vehicle assigned to it."
                    )

            vehicles = (
                session.query(Vehicle).filter(Vehicle.scenario_id == scenario.id).all()
            )
            for vehicle in vehicles:
                if (
                    session.query(Event).filter(Event.vehicle_id == vehicle.id).count()
                    == 0
                ):
                    # Also add a dummy standby-departure event if this vehicle has no events
                    rotation_per_vehicle = sorted(
                        vehicle.rotations, key=lambda r: r.trips[0].departure_time
                    )
                    earliest_trip = rotation_per_vehicle[0].trips[0]
                    area = (
                        session.query(Area)
                        .filter(Area.scenario_id == scenario.id)
                        .filter(Area.vehicle_type_id == Vehicle.vehicle_type_id)
                        .first()
                    )

                    standby_start = earliest_trip.departure_time - timedelta(seconds=1)
                    standby_event = Event(
                        scenario_id=scenario.id,
                        vehicle_type_id=vehicle.vehicle_type_id,
                        vehicle=vehicle,
                        area_id=area.id,
                        subloc_no=area.capacity,
                        time_start=standby_start,
                        time_end=earliest_trip.departure_time,
                        soc_start=1,
                        soc_end=1,
                        event_type=EventType.STANDBY_DEPARTURE,
                        description=f"DUMMY Initial standby event for rotation {earliest_trip.rotation_id}.",
                        timeseries=None,
                    )
                    session.add(standby_event)

        # Since we are doing no_autoflush blocks later, we need to flush the session once here so that unflushed stuff
        # From preceding functions is visible in the database
        session.flush()

        for rotation in rotations:
            rotation: Rotation
            with session.no_autoflush:
                vehicle_type = rotation.vehicle_type
                vehicle = rotation.vehicle
            if vehicle_type.consumption is None:
                raise ValueError(
                    "The vehicle type does not have a consumption value set."
                )
            consumption = vehicle_type.consumption

            # The departure SoC for this rotation is the SoC of the last event preceding the first trip
            with session.no_autoflush:
                current_soc = (
                    session.query(Event.soc_end)
                    .filter(Event.vehicle_id == rotation.vehicle_id)
                    .filter(Event.time_end <= rotation.trips[0].departure_time)
                    .order_by(Event.time_end.desc())
                    .first()[0]
                )

            for trip in rotation.trips:
                # Set up a timeseries
                soc_start = current_soc
                if calculate_timeseries and len(trip.stop_times) > 0:
                    timeseries = {
                        "time": [],
                        "soc": [],
                        "distance": [],
                    }
                    for i in range(len(trip.stop_times)):
                        current_time = trip.stop_times[i].arrival_time
                        dwell_duration = trip.stop_times[i].dwell_duration
                        elapsed_distance = trip.route.assoc_route_stations[
                            i
                        ].elapsed_distance
                        elapsed_energy = consumption * (elapsed_distance / 1000)  # kWh
                        soc = (
                            current_soc - elapsed_energy / vehicle_type.battery_capacity
                        )
                        timeseries["time"].append(current_time.isoformat())
                        timeseries["soc"].append(soc)
                        timeseries["distance"].append(elapsed_distance)
                        if dwell_duration > timedelta(seconds=0):
                            timeseries["time"].append(
                                (current_time + dwell_duration).isoformat()
                            )
                            timeseries["soc"].append(soc)
                            timeseries["distance"].append(elapsed_distance)
                else:
                    timeseries = None
                energy_used = consumption * trip.route.distance / 1000  # kWh
                current_soc = soc_start - energy_used / vehicle_type.battery_capacity

                # Create a driving event
                current_event = Event(
                    scenario_id=scenario.id,
                    vehicle_type_id=rotation.vehicle_type_id,
                    vehicle=vehicle,
                    trip_id=trip.id,
                    time_start=trip.departure_time,
                    time_end=trip.arrival_time,
                    soc_start=soc_start,
                    soc_end=current_soc,
                    event_type=EventType.DRIVING,
                    description=f"`VehicleType.consumption`-based driving event for trip {trip.id}.",
                    timeseries=timeseries,
                )
                session.add(current_event)

                # If the vehicle is
                #  - Capable of opportunity charging
                #  - On a Rotation which allows opportunity charging
                #  - Currently at a station which allows opportunity charging
                #  - which is not the last trip of the rotation
                #  We add a charging event

                if (
                    rotation.vehicle_type.opportunity_charging_capable
                    and rotation.allow_opportunity_charging
                    and trip.route.arrival_station.is_electrified
                    and trip != rotation.trips[-1]
                ):
                    logger.debug(
                        f"Adding opportunity charging event for trip {trip.id}"
                    )
                    # Identify the break time between trips
                    trip_index = rotation.trips.index(trip)
                    next_trip = rotation.trips[trip_index + 1]
                    break_time = next_trip.departure_time - trip.arrival_time

                    # How much energy can be charged in this time?
                    energy_charged = (
                        max([v[1] for v in vehicle_type.charging_curve])
                        * (
                            break_time.total_seconds()
                            - terminus_deadtime.total_seconds()
                        )
                        / 3600
                    )

                    if energy_charged > 0:
                        # Calculate the end SoC
                        post_charge_soc = min(
                            current_soc
                            + energy_charged / vehicle_type.battery_capacity,
                            1,
                        )

                        # Create a simple timeseries for the charging event
                        timeseries = {
                            "time": [
                                trip.arrival_time.isoformat(),
                                (trip.arrival_time + terminus_deadtime / 2).isoformat(),
                                (
                                    next_trip.departure_time - terminus_deadtime / 2
                                ).isoformat(),
                                next_trip.departure_time.isoformat(),
                            ],
                            "soc": [
                                current_soc,
                                current_soc,
                                post_charge_soc,
                                post_charge_soc,
                            ],
                        }

                        # Create the charging event
                        current_event = Event(
                            scenario_id=scenario.id,
                            vehicle_type_id=rotation.vehicle_type_id,
                            vehicle=vehicle,
                            station_id=trip.route.arrival_station_id,
                            time_start=trip.arrival_time,
                            time_end=next_trip.departure_time,
                            soc_start=current_soc,
                            soc_end=post_charge_soc,
                            event_type=EventType.CHARGING_OPPORTUNITY,
                            description=f"Opportunity charging event for trip {trip.id}.",
                            timeseries=timeseries,
                        )
                        current_soc = post_charge_soc
                        session.add(current_event)


def generate_depot_layout(
    scenario: Union[Scenario, int, Any],
    charging_power: float = 150,
    database_url: Optional[str] = None,
    delete_existing_depot: bool = False,
):
    """
    Generates one or more depots for the scenario.

    First, the rotations are scanned to identify all the spots that serve as start *and* end of a rotation. Then the set
    of rotations for these spots are checked for vehicle types that are used there. Next, the amount of vehicles that
    are simultaneously present at the depot is calculated. Then a depot layout with an arrival and a charging area for
    each vehicle type is created. The capacity of each area is taken from the calculated amount of vehicles.
    The depot layout is then added to the database.

    A default plan will also be generated, which includes the following default processes: standby_arrival, cleaning,
    charging and standby_departure. Each vehicle will be processed with this exact order (standby_arrival is optional
    because it only happens if a vehicle needs to wait for the next process).

    The function only deletes the depot if the `delete_existing_depot` parameter is set to True. If there is already a
    depot existing in this scenario and this parameter is set to False, a ValueError will be raised.

    :param scenario: Either a :class:`eflips.model.Scenario` object containing the input data for the simulation. Or
        an integer specifying the ID of a scenario in the database. Or any other object that has an attribute
        ``id`` that is an integer. If no :class:`eflips.model.Scenario` object is passed, the ``database_url``
        parameter must be set to a valid database URL ot the environment variable ``DATABASE_URL`` must be set to a
        valid database URL.

    :param charging_power: the charging power of the charging area in kW

    :param delete_existing_depot: if there is already a depot existing in this scenario, set True to delete this
        existing depot. Set to False and a ValueError will be raised if there is a depot in this scenario.

    :return: None. The depot layout will be added to the database.
    """
    CLEAN_DURATION = 30 * 60  # 30 minutes in seconds

    with create_session(scenario, database_url) as (session, scenario):
        # Handles existing depot
        if session.query(Depot).filter(Depot.scenario_id == scenario.id).count() != 0:
            if delete_existing_depot is False:
                raise ValueError("Depot already exists.")
            delete_depots(scenario, session)

        # Identify all the spots that serve as start *and* end of a rotation
        for (
            first_last_stop_tup,
            vehicle_type_dict,
        ) in group_rotations_by_start_end_stop(scenario.id, session).items():
            first_stop, last_stop = first_last_stop_tup
            if first_stop != last_stop:
                raise ValueError("First and last stop of a rotation are not the same.")
            max_occupancies: Dict[eflips.model.VehicleType, int] = {}
            max_clean_occupancies: Dict[eflips.model.VehicleType, int] = {}
            for vehicle_type, rotations in vehicle_type_dict.items():
                # Slightly convoluted vehicle summation
                start_time = min(
                    [rotation.trips[0].departure_time for rotation in rotations]
                ).timestamp()
                end_time = max(
                    [rotation.trips[-1].arrival_time for rotation in rotations]
                ).timestamp()
                timestamps_to_sample = np.arange(start_time, end_time, 60)
                occupancy = np.zeros_like(timestamps_to_sample)
                clean_occupancy = np.zeros_like(timestamps_to_sample)
                for rotation in rotations:
                    rotation_start = rotation.trips[0].departure_time.timestamp()
                    rotation_end = rotation.trips[-1].arrival_time.timestamp()
                    occupancy += np.interp(
                        timestamps_to_sample,
                        [rotation_start, rotation_end],
                        [1, 1],
                        left=0,
                        right=0,
                    )
                    clean_occupancy += np.interp(
                        timestamps_to_sample,
                        [rotation_end, rotation_end + CLEAN_DURATION],
                        [1, 1],
                        left=0,
                        right=0,
                    )
                max_occupancies[vehicle_type] = max(
                    max(occupancy), 1
                )  # To avoid zero occupancy
                max_clean_occupancies[vehicle_type] = max(max(clean_occupancy), 1)

            # Create a simple depot at this station
            create_simple_depot(
                scenario=scenario,
                station=first_stop,
                charging_capacities=max_occupancies,
                cleaning_capacities=max_clean_occupancies,
                charging_power=charging_power,
                session=session,
                cleaning_duration=timedelta(seconds=CLEAN_DURATION),
                safety_margin=0.2,
            )


def apply_even_smart_charging(
    scenario: Union[Scenario, int, Any],
    database_url: Optional[str] = None,
    standby_departure_duration: timedelta = timedelta(minutes=5),
) -> None:
    """
    Takes a scenario where depot simulation has been run and applies smart charging to the depot.

    This modifies the time and power of the charging events in the database. The arrival and departure times and SoCs at
    these times are not modified.

    :param scenario: A :class:`eflips.model.Scenario` object containing the input data for the simulation.

    :param database_url: An optional database URL. If no database URL is passed and the `scenario` parameter is not a
        :class:`eflips.model.Scenario` object, the environment variable `DATABASE_URL` must be set to a valid database
        URL.

    :param standby_departure_duration: The duration of the STANDBY_DEPARTURE event. This is the time the vehicle is
        allowed to wait at the depot before it has to leave. The default is 5 minutes.

    :return: None. The results are added to the database.
    """
    with create_session(scenario, database_url) as (session, scenario):
        depots = session.query(Depot).filter(Depot.scenario_id == scenario.id).all()
        for depot in depots:
            # Load all the charging events at this depot
            charging_events = (
                session.query(Event)
                .join(Area)
                .filter(Area.depot_id == depot.id)
                .filter(Event.event_type == EventType.CHARGING_DEPOT)
                .all()
            )

            # For each event, take the subsequent STANDBY_DEPARTURE event of the same vehicle
            # Reduce the STANDBY_DEPARTURE events duration to 5 minutes
            # Move the end time of the charging event to the start time of the STANDBY_DEPARTURE event
            for charging_event in charging_events:
                next_event = (
                    session.query(Event)
                    .filter(Event.time_start >= charging_event.time_end)
                    .filter(Event.vehicle_id == charging_event.vehicle_id)
                    .order_by(Event.time_start)
                    .first()
                )

                if (
                    next_event is None
                    or next_event.event_type != EventType.STANDBY_DEPARTURE
                ):
                    warnings.warn(
                        f"Event {charging_event.id} has no STANDBY_DEPARTURE event after a CHARGING_DEPOT "
                        f"event. No room for smart charging."
                    )
                    continue

                assert next_event.time_start == charging_event.time_end

                if (
                    next_event.time_end - next_event.time_start
                ) > standby_departure_duration:
                    next_event.time_start = (
                        next_event.time_end - standby_departure_duration
                    )
                    session.flush()
                    # Add a timeseries to the charging event
                    assert charging_event.timeseries is None
                    charging_event.timeseries = {
                        "time": [
                            charging_event.time_start.isoformat(),
                            charging_event.time_end.isoformat(),
                            next_event.time_start.isoformat(),
                        ],
                        "soc": [
                            charging_event.soc_start,
                            charging_event.soc_end,
                            charging_event.soc_end,
                        ],
                    }
                    charging_event.time_end = next_event.time_start
                    session.flush()

            optimize_charging_events_even(charging_events)


def simulate_scenario(
    scenario: Union[Scenario, int, Any],
    repetition_period: Optional[timedelta] = None,
    database_url: Optional[str] = None,
    smart_charging_strategy: SmartChargingStrategy = SmartChargingStrategy.EVEN,
) -> None:
    """
    This method simulates a scenario and adds the results to the database.

    It fills in the "Charging Events" in the :class:`eflips.model.Event` table and associates
    :class:`eflips.model.Vehicle` objects with all the existing "Driving Events" in the :class:`eflips.model.Event`
    table.

    :param scenario: Either a :class:`eflips.model.Scenario` object containing the input data for the simulation. Or
        an integer specifying the ID of a scenario in the database. Or any other object that has an attribute
        ``id`` that is an integer. If no :class:`eflips.model.Scenario` object is passed, the ``database_url``
        parameter must be set to a valid database URL ot the environment variable ``DATABASE_URL`` must be set to a
        valid database URL.

    :param repetition_period: An optional timedelta object specifying the period of the vehicle schedules. This
        is needed because the result should be a steady-state result. THis can only be achieved by simulating a
        time period before and after our actual simulation, and then only using the "middle". eFLIPS tries to
        automatically detect whether the schedule should be repeated daily or weekly. If this fails, a ValueError is
        raised and repetition needs to be specified manually.

    :param database_url: An optional database URL. If no database URL is passed and the `scenario` parameter is not a
        :class:`eflips.model.Scenario` object, the environment variable `DATABASE_URL` must be set to a valid database
        URL.

    :param smart_charging_strategy: An optional parameter specifying the smart charging strategy to be used. The
        default is SmartChargingStragegy.NONE. The following strategies are available:
        - SmartChargingStragegy.NONE: Do not use smart charging. Buses are charged with the maximum power available,
        from the time they arrive at the depot until they are full (or leave the depot).
        - SmartChargingStragegy.EVEN: Use smart charging with an even distribution of charging power over the time the
        bus is at the depot. This aims to minimize the peak power demand.

    :return: Nothing. The results are added to the database.
    """
    with create_session(scenario, database_url) as (session, scenario):
        simulation_host = init_simulation(
            scenario=scenario,
            session=session,
            repetition_period=repetition_period,
        )
        ev = run_simulation(simulation_host)
        add_evaluation_to_database(scenario, ev, session)

        match smart_charging_strategy:
            case SmartChargingStrategy.NONE:
                pass
            case SmartChargingStrategy.EVEN:
                apply_even_smart_charging(scenario, database_url)
            case SmartChargingStrategy.MIN_PRICE:
                raise NotImplementedError("MIN_PRICE strategy is not implemented yet.")
            case _:
                raise NotImplementedError()


def _init_simulation(
    scenario: Scenario,
    session: Session,
    repetition_period: Optional[timedelta] = None,
    vehicle_count_dict: Optional[Dict[str, int]] = None,
) -> SimulationHost:
    """Deprecated stub for init_simulation."""
    raise NotImplementedError(
        "The function _init_simulation is deprecated. Please use init_simulation instead."
    )


def init_simulation(
    scenario: Scenario,
    session: Session,
    repetition_period: Optional[timedelta] = None,
    vehicle_count_dict: Optional[Dict[str, Dict[str, int]]] = None,
) -> SimulationHost:
    """
    This methods checks the input data for consistency, initializes a simulation host object and returns it.

    The simulation host object can then be passed to :func:`run_simulation()`.

    :param scenario: A :class:`eflips.model.Scenario` object containing the input data for the simulation.

    :param session: A SQLAlchemy session object.

    :param repetition_period: An optional timedelta object specifying the period of the vehicle schedules. This
        is needed because the *result* should be a steady-state result. THis can only be achieved by simulating a
        time period before and after our actual simulation, and then only using the "middle". eFLIPS tries to
        automatically detect whether the schedule should be repeated daily or weekly. If this fails, a ValueError is
        raised and repetition needs to be specified manually.

    :param vehicle_count_dict: An optional dictionary specifying the number of vehicles for each vehicle type for each
        depot. The dictionary should have the following structure:

         ::

            {
                "1" (depot.id as str): {
                    "1" (vehicle_type.id as str): 10,
                    "2" (vehicle_type.id as str): 20,
                    ...
                },
                "2" (depot.id as str): {
                    "1" (vehicle_type.id as str): 10,
                    "2" (vehicle_type.id as str): 20,
                    ...
                },

    :return: A :class:`eflips.depot.Simulation.SimulationHost` object. This object should be reagrded as a "black box"
        by the user. It should be passed to :func:`run_simulation()` to run the simulation and obtain the results.
    """
    # Clear the eflips settings
    eflips.settings.reset_settings()

    path_to_this_file = os.path.dirname(__file__)

    # Step 1: Set up the depot
    eflips_depots = []
    for depot in scenario.depots:
        # Step 1.5: Check validity of a depot
        check_depot_validity(depot)

        depot_dict = depot_to_template(depot)
        eflips_depots.append(
            eflips.depot.Depotinput(filename_template=depot_dict, show_gui=False)
        )

    # Step 2: eFLIPS initialization
    simulation_host = SimulationHost(eflips_depots, print_timestamps=False)
    # Now we do what is done in `standard_setup()` from the old input files
    # Load the settings
    path_to_default_settings = os.path.join(
        path_to_this_file, "defaults", "default_settings"
    )
    eflips.load_settings(path_to_default_settings)

    # Step 3: Set up the vehicle schedules
    # Turn rotations into vehicleschedules
    # Turn API VehicleSchedule objects into eFLIPS TimeTable object
    # Get correctly repeated vehicle schedules
    # if total duration time is 1 or 2 days, vehicle schedule will be repeated daily
    # if total duration time is 7 or 8 days, vehicle schedule will be repeated weekly
    # However, we need to override quite a lot of the settings
    # The ["general"]["SIMULATION_TIME"] entry is calculated from the difference between the first and last departure
    # time in the vehicle schedule
    vehicle_schedules = [
        VehicleSchedule.from_rotation(
            rotation,
            scenario=scenario,
            session=session,
        )
        for rotation in scenario.rotations
    ]

    first_departure_time = min(
        [vehicle_schedule.departure for vehicle_schedule in vehicle_schedules]
    )
    last_arrival_time = max(
        [vehicle_schedule.arrival for vehicle_schedule in vehicle_schedules]
    )

    # We take first arrival time as simulation start
    total_duration = (last_arrival_time - first_departure_time).total_seconds()
    schedule_duration_days = ceil(total_duration / (24 * 60 * 60))

    if repetition_period is None and schedule_duration_days in [1, 2]:
        repetition_period = timedelta(days=1)
    elif repetition_period is None and schedule_duration_days in [7, 8]:
        repetition_period = timedelta(weeks=1)
    elif repetition_period is None:
        raise ValueError(
            "Could not automatically detect repetition period. Please specify manually."
        )

    # Now, we need to repeat the vehicle schedules

    vehicle_schedules = repeat_vehicle_schedules(vehicle_schedules, repetition_period)

    sim_start_stime, total_duration_seconds = start_and_end_times(vehicle_schedules)
    eflips.globalConstants["general"]["SIMULATION_TIME"] = int(total_duration_seconds)

    eflips.globalConstants["general"]["SIMULATION_START_DATETIME"] = sim_start_stime

    timetable = VehicleSchedule._to_timetable(
        vehicle_schedules, simulation_host.env, sim_start_stime
    )
    simulation_host.timetable = timetable

    # Step 4: Set up the vehicle types
    # We need to calculate roughly how many vehicles we need for each depot
    for depot in scenario.depots:
        depot_id = str(depot.id)
        eflips.globalConstants["depot"]["vehicle_count"][depot_id] = {}
        vehicle_types_for_depot = set(str(area.vehicle_type_id) for area in depot.areas)
        if "None" in vehicle_types_for_depot:
            vehicle_types_for_depot.remove("None")

        # If we have a vehicle count dictionary, we validate and use ir
        if vehicle_count_dict is not None and depot_id in vehicle_count_dict.keys():
            if set(vehicle_count_dict[depot_id].keys()) < vehicle_types_for_depot:
                raise ValueError(
                    "The vehicle count dictionary does not contain all vehicle types for depot {depot_id}."
                )
            eflips.globalConstants["depot"]["vehicle_count"][
                depot_id
            ] = vehicle_count_dict[depot_id]
        else:
            # Calculate it from the size of the charging area with a 2x margin

            for vehicle_type in vehicle_types_for_depot:
                vehicle_count = 0
                for area in depot.areas:
                    if area.vehicle_type_id == int(vehicle_type):
                        # TODO potential edit if we make vehicle type of an area a list
                        for p in area.processes:
                            if p.electric_power is not None and p.duration is None:
                                vehicle_count = area.capacity

                assert (
                    vehicle_count > 0
                ), f"The charging area capacity for vehicle type {vehicle_type} should not be 0."

                eflips.globalConstants["depot"]["vehicle_count"][depot_id][
                    vehicle_type
                ] = (vehicle_count * 2)

    # We  need to put the vehicle type objects into the GlobalConstants
    for vehicle_type in scenario.vehicle_types:
        eflips.globalConstants["depot"]["vehicle_types"][
            str(vehicle_type.id)
        ] = vehicle_type_to_global_constants_dict(vehicle_type)

    # Step 5: Final checks and setup
    # Run the eflips validity checks
    eflips.depot.settings_config.check_gc_validity()

    # Complete the eflips settings
    eflips.depot.settings_config.complete_gc()

    # Set up the depots
    for dh, di in zip(simulation_host.depot_hosts, simulation_host.to_simulate):
        dh.load_and_complete_template(di.filename_template)

    simulation_host.complete()

    return simulation_host


def _run_simulation(simulation_host: SimulationHost) -> DepotEvaluation:
    """Deprecated stub for run_simulation."""
    raise NotImplementedError(
        "The function _run_simulation is deprecated. Please use run_simulation instead."
    )


def run_simulation(simulation_host: SimulationHost) -> Dict[str, DepotEvaluation]:
    """Run simulation and return simulation results.

    :param simulation_host: A "black box" object containing all input data for the simulation.

    :return: A dictionary of :class:`eflips.depot.evaluation.DepotEvaluation` objects. The keys are the depot IDs, as
        strings.
    """
    simulation_host.run()

    results = {}
    for depot_host in simulation_host.depot_hosts:
        depot_id = depot_host.depot.ID
        ev = depot_host.evaluation

        # We need to clean up the timetable, it has trips from all depots
        ev.timetable = copy.copy(ev.timetable)
        ev.timetable.trips = [
            t for t in ev.timetable.trips if t.destination.ID == depot_id
        ]
        ev.timetable.trips_issued = [
            t for t in ev.timetable.trips_issued if t.destination.ID == depot_id
        ]
        ev.timetable.all_trips = [
            t for t in ev.timetable.all_trips if t.destination.ID == depot_id
        ]

        # We also need to clean up the vehicle generator
        ev.vehicle_generator = copy.copy(ev.vehicle_generator)
        ev.vehicle_generator.items = copy.copy(ev.vehicle_generator.items)
        indizes_to_remove = []
        for i in range(len(ev.vehicle_generator.items)):
            vehicle = copy.copy(ev.vehicle_generator.items[i])
            depot_ids = [trip.destination.ID for trip in vehicle.finished_trips]
            if len(set(depot_ids)) > 1:
                raise ValueError("Vehicle has finished trips in multiple depots.")
            if len(depot_ids) == 0 or depot_ids[0] != depot_id:
                indizes_to_remove.append(i)

        for index in sorted(indizes_to_remove, reverse=True):
            del ev.vehicle_generator.items[index]

        results[depot_id] = ev

    return results


def _add_evaluation_to_database(
    scenario_id: int,
    depot_evaluation: DepotEvaluation,
    session: sqlalchemy.orm.Session,
) -> None:
    """Deprecated stub for add_evaluation_to_database."""
    raise NotImplementedError(
        "The function _add_evaluation_to_database is deprecated. Please use add_evaluation_to_database instead."
    )


def add_evaluation_to_database(
    scenario: Scenario,
    depot_evaluations: Dict[str, DepotEvaluation],
    session: sqlalchemy.orm.Session,
) -> None:
    """
    This method adds a simulation result to the database.

    It reads the simulation results from the :class:`eflips.depot.evaluation.DepotEvaluation` object and  adds them into
     the database. Tables of Event, Rotation and Vehicle will be updated.

    :param scenario: A :class:`eflips.model.Scenario` object containing the input data for the simulation.

    :param depot_evaluations: A dictionary of :class:`eflips.depot.evaluation.DepotEvaluation` objects. The keys are
        the depot IDs, as strings.

    :param session: a SQLAlchemy session object. This is used to add all the simulation results to the
        database.

    :return: Nothing. The results are added to the database.
    """

    # Read simulation start time

    for depot_id, depot_evaluation in depot_evaluations.items():
        simulation_start_time = depot_evaluation.sim_start_datetime

        # Depot-layer operations

        list_of_assigned_schedules = []

        waiting_area_id = None

        total_areas = scenario.areas
        for area in total_areas:
            if area.depot_id == int(depot_id) and len(area.processes) == 0:
                waiting_area_id = area.id

        assert isinstance(waiting_area_id, int) and waiting_area_id > 0, (
            f"Waiting area id should be an integer greater than 0. For every depot there must be at least "
            f"one waiting area."
        )

        for current_vehicle in depot_evaluation.vehicle_generator.items:
            # Vehicle-layer operations

            vehicle_type_id = int(current_vehicle.vehicle_type.ID)

            current_vehicle_db = Vehicle(
                vehicle_type_id=vehicle_type_id,
                scenario=scenario,
                name=current_vehicle.ID,
                name_short=None,
            )

            # Flush the vehicle object to get the vehicle id
            session.add(current_vehicle_db)
            session.flush()

            dict_of_events = OrderedDict()

            (
                schedule_current_vehicle,
                earliest_time,
                latest_time,
                # Earliest and latest time defines a time window, only the events within this time window will be
                # handled. It is usually the departure time of the last copy trip in the "early-shifted" copy
                # schedules and the departure time of the first copy trip in the "late-shifted" copy schedules.
            ) = _get_finished_schedules_per_vehicle(
                dict_of_events, current_vehicle.finished_trips, current_vehicle_db.id
            )

            try:
                assert earliest_time is not None and latest_time is not None

            except AssertionError as e:
                warnings.warn(
                    f"Vehicle {current_vehicle_db.id} has only copied trips. The profiles of this vehicle "
                    f"will not be written into database."
                )
                continue

            assert (
                earliest_time < latest_time
            ), f"Earliest time {earliest_time} is not less than latest time {latest_time}."

            list_of_assigned_schedules.extend(schedule_current_vehicle)

            _generate_vehicle_events(
                dict_of_events,
                current_vehicle,
                waiting_area_id,
                earliest_time,
                latest_time,
            )

            # Python passes dictionaries by reference

            _complete_standby_departure_events(dict_of_events, latest_time)

            _add_soc_to_events(dict_of_events, current_vehicle.battery_logs)

            try:
                assert (not dict_of_events) is False
            except AssertionError as e:
                warnings.warn(
                    f"Vehicle {current_vehicle_db.id} has no valid events. The vehicle will not be written "
                    f"into database."
                )

                continue

            _add_events_into_database(
                current_vehicle_db,
                dict_of_events,
                session,
                scenario,
                simulation_start_time,
            )

        # Postprocessing of events
        _update_vehicle_in_rotation(session, scenario, list_of_assigned_schedules)
        _update_waiting_events(session, scenario, waiting_area_id)


def _get_finished_schedules_per_vehicle(
    dict_of_events, list_of_finished_trips: List, db_vehicle_id: int
):
    """
    This function completes the following tasks:

    1. It gets the finished non-copy schedules of the current vehicle,
    which will be used in :func:`_update_vehicle_in_rotation()`.

    2. It fills the dictionary of events with the trip_ids of the current vehicle.

    3. It returns an earliest and a latest time according to this vehicle's schedules. Only processes happening within
    this time window will be handled later.

    Usually the earliest time is the departure time of the last copy trip in the "early-shifted" copy schedules
    and the lastest time is the departure time of the first copy trip in the "late-shifted" copy schedules.

    # If the vehicle's first trip is a non-copy trip, the earliest time is the departure time of the first trip. If the
    # vehicle's last trip is a non-copy trip, the latest time is the departure time of the last trip.

    :param dict_of_events: An ordered dictionary storing the data related to an event. The keys are the start times of
        the events.
    :param list_of_finished_trips: A list of finished trips of a vehicle directly from
        :class:`eflips.depot.simple_vehicle.SimpleVehicle` object.

    :param db_vehicle_id: The vehicle id in the database.

    :return: A tuple of three elements. The first element is a list of finished schedules of the vehicle. The second and
        third elements are the earliest and latest time of the vehicle's schedules.
    """
    finished_schedules = []

    list_of_finished_trips.sort(key=lambda x: x.atd)
    earliest_time = None
    latest_time = None

    for i in range(len(list_of_finished_trips)):
        assert list_of_finished_trips[i].atd == list_of_finished_trips[i].std, (
            "The trip {current_trip.ID} is delayed. The simulation doesn't "
            "support delayed trips for now."
        )

        if list_of_finished_trips[i].is_copy is False:
            current_trip = list_of_finished_trips[i]

            finished_schedules.append((int(current_trip.ID), db_vehicle_id))
            dict_of_events[current_trip.atd] = {
                "type": "Trip",
                "id": int(current_trip.ID),
            }
            if i == 0:
                raise ValueError(
                    f"New Vehicle required for the trip {current_trip.ID}, which suggests the fleet or the "
                    f"infrastructure might not be enough for the full electrification. Please add charging "
                    f"interfaces or increase charging power ."
                )

            elif i != 0 and i == len(list_of_finished_trips) - 1:
                # Vehicle's last trip is a non-copy trip
                if earliest_time is None:
                    earliest_time = list_of_finished_trips[i - 1].ata
                latest_time = list_of_finished_trips[i].ata

            else:
                if list_of_finished_trips[i - 1].is_copy is True:
                    earliest_time = list_of_finished_trips[i - 1].ata
                if list_of_finished_trips[i + 1].is_copy is True:
                    latest_time = list_of_finished_trips[i + 1].atd

    return finished_schedules, earliest_time, latest_time


def _generate_vehicle_events(
    dict_of_events,
    current_vehicle: SimpleVehicle,
    virtual_waiting_area_id: int,
    earliest_time: datetime.datetime,
    latest_time: datetime.datetime,
) -> None:
    """
    This function generates and ordered dictionary storing the data related to an event.

    It returns a dictionary. The keys are the start times of the
    events. The values are also dictionaries containing:
    - type: The type of the event.
    - end: The end time of the event.
    - area: The area id of the event.
    - slot: The slot id of the event.
    - id: The id of the event-related process.

    For trips, only the type is stored.

    For waiting events, the slot is not stored for now.

    :param current_vehicle: a :class:`eflips.depot.simple_vehicle.SimpleVehicle` object.

    :param virtual_waiting_area_id: the id of the virtual waiting area. Vehicles waiting for the first process will park here.

    :param earliest_time: the earliest relevant time of the current vehicle. Any events earlier than this will not be
        handled.

    :param latest_time: the latest relevant time of the current vehicle. Any events later than this will not be handled.

    :return: None. The results are added to the dictionary.
    """

    # For convenience
    area_log = current_vehicle.logger.loggedData["dwd.current_area"]
    slot_log = current_vehicle.logger.loggedData["dwd.current_slot"]
    waiting_log = current_vehicle.logger.loggedData["area_waiting_time"]

    # Handling waiting events
    waiting_log_timekeys = sorted(waiting_log.keys())

    for idx in range(len(waiting_log_timekeys)):
        waiting_end_time = waiting_log_timekeys[idx]

        # Only extract events if the time is within the upper mentioned range

        if earliest_time <= waiting_end_time <= latest_time:
            waiting_info = waiting_log[waiting_end_time]

            if waiting_info["waiting_time"] == 0:
                continue

            warnings.warn(
                f"Vehicle {current_vehicle.ID} has been waiting for {waiting_info['waiting_time']} seconds. "
            )

            start_time = waiting_end_time - waiting_info["waiting_time"]

            if waiting_info["area"] == waiting_log[waiting_log_timekeys[0]]["area"]:
                # if the vehicle is waiting for the first process, put it in the virtual waiting area
                waiting_area_id = virtual_waiting_area_id
            else:
                # If the vehicle is waiting for other processes,
                # put it in the area of the prodecessor process of the waited process.
                waiting_area_id = waiting_log[waiting_log_timekeys[idx - 1]]["area"]

            dict_of_events[start_time] = {
                "type": "Standby",
                "end": waiting_end_time,
                "area": waiting_area_id,
                "is_waiting": True,
            }

    # Create a list of battery log in order of time asc. Convenient for looking up corresponding soc

    for time_stamp, process_log in current_vehicle.logger.loggedData[
        "dwd.active_processes_copy"
    ].items():
        if earliest_time <= time_stamp <= latest_time:
            num_process = len(process_log)
            if num_process == 0:
                # A departure happens and this trip should already be stored in the dictionary
                pass
            else:
                for process in process_log:
                    current_area = area_log[time_stamp]
                    current_slot = slot_log[time_stamp]

                    if current_area is None or current_slot is None:
                        raise ValueError(
                            f"For process {process.ID} Area and slot should not be None."
                        )

                    match process.status:
                        case ProcessStatus.COMPLETED | ProcessStatus.CANCELLED:
                            assert (
                                len(process.starts) == 1 and len(process.ends) == 1
                            ), (
                                f"Current process {process.ID} is completed and should only contain one start and "
                                f"one end time."
                            )

                            if process.dur > 0:
                                # Valid duration
                                dict_of_events[time_stamp] = {
                                    "type": type(process).__name__,
                                    "end": process.ends[0],
                                    "area": current_area.ID,
                                    "slot": current_slot,
                                    "id": process.ID,
                                }
                            else:
                                # Duration is 0
                                assert current_area.issink is True, (
                                    f"A process with no duration could only "
                                    f"happen in the last area before dispatched"
                                )
                                if (
                                    time_stamp in dict_of_events.keys()
                                    and "end" in dict_of_events[time_stamp].keys()
                                ):
                                    start_this_event = dict_of_events[time_stamp]["end"]
                                    dict_of_events[start_this_event] = {
                                        "type": type(process).__name__,
                                        "area": current_area.ID,
                                        "slot": current_slot,
                                        "id": process.ID,
                                    }

                        case ProcessStatus.IN_PROGRESS:
                            assert (
                                len(process.starts) == 1 and len(process.ends) == 0
                            ), f"Current process {process.ID} is marked IN_PROGRESS, but has an end."

                            if current_area is None or current_slot is None:
                                raise ValueError(
                                    f"For process {process.ID} Area and slot should not be None."
                                )

                            if process.dur > 0:
                                # Valid duration
                                dict_of_events[time_stamp] = {
                                    "type": type(process).__name__,
                                    "end": process.etc,
                                    "area": current_area.ID,
                                    "slot": current_slot,
                                    "id": process.ID,
                                }
                            else:
                                raise NotImplementedError(
                                    "We believe this should never happen. If it happens, handle it here."
                                )

                        # The following ProcessStatus possibly only happen while the simulation is running,
                        # not in the results
                        case ProcessStatus.WAITING:
                            raise NotImplementedError(
                                f"Current process {process.ID} is waiting. Not implemented yet."
                            )

                        case ProcessStatus.NOT_STARTED:
                            raise NotImplementedError(
                                f"Current process {process.ID} is not started. Not implemented yet."
                            )

                        case _:
                            raise ValueError(
                                f"Invalid process status {process.status} for process {process.ID}."
                            )


def _complete_standby_departure_events(
    dict_of_events: Dict, latest_time: datetime.datetime
) -> None:
    """
    This function completes the standby departure events by adding an end time to each standby departure event.

    :param dict_of_events: a dictionary containing the events of a vehicle. The keys are the start times of the events.

    :param latest_time: the latest relevant time of the current vehicle. Any events later than this will not be handled.

    :return: None. The results are added to the dictionary.
    """
    for i in range(len(dict_of_events.keys())):
        time_keys = sorted(dict_of_events.keys())

        process_dict = dict_of_events[time_keys[i]]
        if "end" not in process_dict and process_dict["type"] != "Trip":
            # End time of a standby_departure will be the start of the following trip
            if i == len(time_keys) - 1:
                # The event reaches simulation end
                end_time = latest_time
            else:
                end_time = time_keys[i + 1]

            process_dict["end"] = end_time


def _add_soc_to_events(dict_of_events, battery_log) -> None:
    """
    This function completes the soc of each event by looking up the battery log.

    :param dict_of_events: a dictionary containing the events of a vehicle. The keys are the start times of the events.

    :param battery_log: a list of battery logs of a vehicle.

    :return: None. The results are added to the dictionary.
    """
    battery_log_list = []
    for log in battery_log:
        battery_log_list.append((log.t, log.energy / log.energy_real))

    time_keys = sorted(dict_of_events.keys())
    for i in range(len(time_keys)):
        # Get soc
        soc_start = None
        soc_end = None
        start_time = time_keys[i]
        process_dict = dict_of_events[time_keys[i]]
        for j in range(len(battery_log_list)):
            # Access the correct battery log according to time since there is only one battery log for each time
            log = battery_log_list[j]

            if process_dict["type"] != "Trip":
                if log[0] == start_time:
                    soc_start = log[1]
                if log[0] == process_dict["end"]:
                    soc_end = log[1]
                if log[0] < start_time < battery_log_list[j + 1][0]:
                    soc_start = log[1]
                if log[0] < process_dict["end"] < battery_log_list[j + 1][0]:
                    soc_end = log[1]

                if soc_start is not None:
                    soc_start = min(soc_start, 1)  # so
                process_dict["soc_start"] = soc_start
                if soc_end is not None:
                    soc_end = min(soc_end, 1)  # soc should not exceed 1
                process_dict["soc_end"] = soc_end

            else:
                continue


def _add_events_into_database(
    db_vehicle, dict_of_events, session, scenario, simulation_start_time
) -> None:
    """
    This function generates :class:`eflips.model.Event` objects from the dictionary of events and adds them into the.

    database.

    :param db_vehicle: vehicle object in the database

    :param dict_of_events: dictionary containing the events of a vehicle. The keys are the start times of the events.

    :param session: a :class:`sqlalchemy.orm.Session` object for database connection.

    :param scenario: the current simulated scenario

    :param simulation_start_time: simulation start time in :class:`datetime.datetime` format

    :return: None. The results are added to the database.
    """
    for start_time, process_dict in dict_of_events.items():
        # Generate EventType
        match process_dict["type"]:
            case "Serve":
                event_type = EventType.SERVICE
            case "Charge":
                event_type = EventType.CHARGING_DEPOT
            case "Standby":
                if (
                    "is_waiting" in process_dict.keys()
                    and process_dict["is_waiting"] is True
                ):
                    event_type = EventType.STANDBY
                else:
                    event_type = EventType.STANDBY_DEPARTURE
            case "Precondition":
                event_type = EventType.PRECONDITIONING
            case "Trip":
                continue
            case _:
                raise ValueError(
                    'Invalid process type %s. Valid process types are "Serve", "Charge", '
                    '"Standby", "Precondition"'
                )

        current_event = Event(
            scenario=scenario,
            vehicle_type_id=db_vehicle.vehicle_type_id,
            vehicle=db_vehicle,
            station_id=None,
            area_id=int(process_dict["area"]),
            subloc_no=int(process_dict["slot"])
            if "slot" in process_dict.keys()
            else 00,
            trip_id=None,
            time_start=timedelta(seconds=start_time) + simulation_start_time,
            time_end=timedelta(seconds=process_dict["end"]) + simulation_start_time,
            soc_start=process_dict["soc_start"]
            if process_dict["soc_start"] is not None
            else process_dict["soc_end"],
            soc_end=process_dict["soc_end"]
            if process_dict["soc_end"] is not None
            else process_dict["soc_start"],  # if only one battery log is found,
            # then this is not an event with soc change
            event_type=event_type,
            description=process_dict["id"] if "id" in process_dict.keys() else None,
            timeseries=None,
        )

        session.add(current_event)


def _update_vehicle_in_rotation(session, scenario, list_of_assigned_schedules) -> None:
    """
    This function updates the vehicle id assigned to the rotations and deletes the events that are not depot events.

    :param session: a :class:`sqlalchemy.orm.Session` object for database connection.
    :param scenario: the current simulated scenario
    :param list_of_assigned_schedules: a list of tuples containing the rotation id and the vehicle id.
    :return: None. The results are added to the database.
    """
    # New rotation assignment
    for schedule_id, vehicle_id in list_of_assigned_schedules:
        # Get corresponding old vehicle id
        session.query(Rotation).filter(Rotation.id == schedule_id).update(
            {"vehicle_id": vehicle_id}, synchronize_session="auto"
        )

    # Delete all non-depot events
    session.query(Event).filter(
        Event.scenario == scenario,
        Event.trip_id.isnot(None) | Event.station_id.isnot(None),
    ).delete(synchronize_session="auto")

    session.flush()

    # Delete all vehicles without rotations
    vehicle_assigned_sq = (
        session.query(Rotation.vehicle_id)
        .filter(Rotation.scenario == scenario)
        .distinct()
        .subquery()
    )

    session.query(Vehicle).filter(Vehicle.scenario == scenario).filter(
        Vehicle.id.not_in(select(vehicle_assigned_sq))
    ).delete()

    session.flush()


def _update_waiting_events(session, scenario, waiting_area_id) -> None:
    """
    This function evaluates the capacity of waiting area and assigns the waiting events to corresponding slots in the.

    waiting area.

    :param session: a :class:`sqlalchemy.orm.Session` object for database connection.

    :param scenario: the current simulated scenario.

    :param waiting_area_id: id of the waiting area.

    :raise ValueError: if the waiting area capacity is less than the peak waiting occupancy.

    :return: None. The results are added to the database.
    """
    # Process all the STANDBY (waiting) events #
    all_waiting_starts = (
        session.query(Event)
        .filter(
            Event.scenario_id == scenario.id,
            Event.event_type == EventType.STANDBY,
            Event.area_id == waiting_area_id,
        )
        .all()
    )

    all_waiting_ends = (
        session.query(Event)
        .filter(
            Event.scenario_id == scenario.id,
            Event.event_type == EventType.STANDBY,
            Event.area_id == waiting_area_id,
        )
        .all()
    )

    assert len(all_waiting_starts) == len(
        all_waiting_ends
    ), f"Number of waiting events starts {len(all_waiting_starts)} is not equal to the number of waiting event ends"

    if len(all_waiting_starts) == 0:
        print(
            "No waiting events found. The depot has enough capacity for waiting. Change the waiting area capacity to 10 as buffer."
        )

        session.query(Area).filter(Area.id == waiting_area_id).update(
            {"capacity": 10}, synchronize_session="auto"
        )

        return

    list_waiting_timestamps = []
    for waiting_start in all_waiting_starts:
        list_waiting_timestamps.append(
            {"timestamp": waiting_start.time_start, "event": (waiting_start.id, 1)}
        )

    for waiting_end in all_waiting_ends:
        list_waiting_timestamps.append(
            {"timestamp": waiting_end.time_end, "event": (waiting_end.id, -1)}
        )

    list_waiting_timestamps.sort(key=lambda x: x["timestamp"])
    start_and_end_records = [wt["event"][1] for wt in list_waiting_timestamps]

    peak_waiting_occupancy = max(list(itertools.accumulate(start_and_end_records)))

    # Assuming that there is only one waiting area in each depot

    waiting_area_id = all_waiting_starts[0].area_id
    waiting_area = session.query(Area).filter(Area.id == waiting_area_id).first()
    if waiting_area.capacity > peak_waiting_occupancy:
        warnings.warn(
            f"Current waiting area capacity {waiting_area.capacity} "
            f"is greater than the peak waiting occupancy. Updating the capacity to {peak_waiting_occupancy}."
        )
        session.query(Area).filter(Area.id == waiting_area_id).update(
            {"capacity": peak_waiting_occupancy}, synchronize_session="auto"
        )
        session.flush()
    elif waiting_area.capacity < peak_waiting_occupancy:
        raise ValueError(
            f"Waiting area capacity is less than the peak waiting occupancy. "
            f"Waiting area capacity: {waiting_area.capacity}, peak waiting occupancy: {peak_waiting_occupancy}."
        )
    else:
        pass

    session.flush()

    # Update waiting slots
    virtual_waiting_area = [None] * peak_waiting_occupancy
    for wt in list_waiting_timestamps:
        # check in
        if wt["event"][1] == 1:
            for i in range(len(virtual_waiting_area)):
                if virtual_waiting_area[i] is None:
                    virtual_waiting_area[i] = wt["event"][0]
                    session.query(Event).filter(Event.id == wt["event"][0]).update(
                        {"subloc_no": i}, synchronize_session="auto"
                    )
                    break
        # check out
        else:
            for i in range(len(virtual_waiting_area)):
                if virtual_waiting_area[i] == wt["event"][0]:
                    current_waiting_event = (
                        session.query(Event).filter(Event.id == wt["event"][0]).first()
                    )
                    assert current_waiting_event.subloc_no == i, (
                        f"Subloc number of the event {current_waiting_event.id} is not equal to the index of the "
                        f"event in the virtual waiting area."
                    )
                    virtual_waiting_area[i] = None
                    break

    session.flush()



"""Initializers for population and attributes. Related helpers and constraints.
"""
from operator import attrgetter
from collections import Counter
import random
from copy import deepcopy
from abc import ABC
from deap import creator
from eflips.depot.layout_opt import packing
from eflips.depot.layout_opt.settings import OPT_CONSTANTS as OC


# Functions for precomputing


def get_capacity_max(visu_class, capacity_min, limit=200):
    """Determine an area type's maximum capacity within DEPOT_A, DEPOT_B (the
    b-dimension). Uses BinWithDistances.try_put as a lightweight packing test.
    Return the capacity [int] or None if not even the minimum fits.
    """
    dims = packing.BinWithDistances(
        OC["scenario"]["DEPOT_A"], OC["scenario"]["DEPOT_B"]
    )
    for c in range(capacity_min, limit + 1):
        candidate = visu_class(capacity=c)
        success, _ = dims.try_put(candidate)
        if not success:
            if c == capacity_min:
                # Not even the minimum fits
                return None
            else:
                # Previous fits
                return c - 1
    raise RuntimeError(
        "Determination of the max capacity for %s exceeded the "
        "limit of %d areas." % (visu_class, limit)
    )


def get_count_max(visu_class, capacity):
    """Determine the maximum number of areas of *visu_class* with *capacity*
    within DEPOT_A, DEPOT_B. Return the count [int] or None if not even one
    area fits and the BinWithDistances object populated with the max count.
    """
    dvisu = packing.BinWithDistances(
        OC["scenario"]["DEPOT_A"], OC["scenario"]["DEPOT_B"]
    )
    item = visu_class(capacity=capacity)
    dvisu.items.append(item)
    dvisu.pack()
    while dvisu.feasible:
        item = visu_class(capacity=capacity)
        dvisu.items.append(item)
        dvisu.repack()

    if len(dvisu.items) == 1:
        # Not even 1 area fits
        return None, dvisu
    else:
        # Remove the last item as it didn't fit
        dvisu.items.remove(item)
        # dvisu.repack()    # skipped, dvisu.feasible only seems False until calling repack()
        return len(dvisu.items), dvisu


def get_count_max_with_capacity_max(visu_class, capacity_min):
    """Determine the maximum number of areas of *visu_class* with maximum
    capacity within DEPOT_A, DEPOT_B (the a-dimension). Return the count [int]
    or None if not even one area fits and the BinWithDistances object populated
    with the max count.
    """
    capacity_max = get_capacity_max(visu_class, capacity_min)
    return get_count_max(visu_class, capacity_max)


def get_count_max_with_capacity_min(visu_class, capacity_min):
    """Determine the maximum number of areas of *visu_class* with minimum
    capacity within DEPOT_A, DEPOT_B. Return the count [int] or None if not
    even one area fits and the BinWithDistances object populated with the max
    count.
    """
    return get_count_max(visu_class, capacity_min)


# Initializers


class AreaPrototype(ABC):
    """Abstract base class for a simple area representation."""

    capacity_min = int()
    capacity_max = int()
    typename = str()
    visu_class = None

    def __init__(self, capacity):
        self.capacity = capacity
        self._visu = None  # take care: needs reset after mutation

    def __repr__(self):
        return "{%s} cap=%s" % (self.typename, self.capacity)

    @property
    def visu(self):
        if self._visu is None:
            self._visu = self.visu_class(capacity=self.capacity)
        return self._visu

    @visu.setter
    def visu(self, value):
        self._visu = value

    def to_other_type(self, other_type):
        """Split self. into the least amount of area prototype objects of
        *other_type* with a total capacity of at least self.capacity. Return a
        list of the generated objects.

        other_type: type of other area prototype
        """
        if other_type is type(self):
            return ValueError("Cannot convert to the same type.")

        result = []

        q = self.capacity // other_type.capacity_max
        for i in range(q):
            result.append(other_type(capacity=other_type.capacity_max))

        r = self.capacity % other_type.capacity_max
        if r:
            if r >= other_type.capacity_min:
                result.append(other_type(capacity=r))
            else:
                # unexact return value: capacity is increased to the mininmum
                result.append(other_type(capacity=other_type.capacity_min))

        return result

    def clone(self):
        """Return a new instance with same capacity and no visu."""
        new = type(self)(self.capacity)
        return new


class DSRPrototype(AreaPrototype):
    """Class for a simple DSR area representation."""

    capacity_min = 1
    capacity_max = get_capacity_max(packing.VisuDataDirectSingleRow, capacity_min)
    typename = "DSR"
    visu_class = packing.VisuDataDirectSingleRow

    def __init__(self, capacity):
        super(DSRPrototype, self).__init__(capacity)


class DSR_90Prototype(AreaPrototype):
    """Class for a simple DSR_90 area representation."""

    capacity_min = 1
    capacity_max = get_capacity_max(packing.VisuDataDirectSingleRow_90, capacity_min)
    typename = "DSR_90"
    visu_class = packing.VisuDataDirectSingleRow_90

    def __init__(self, capacity):
        super(DSR_90Prototype, self).__init__(capacity)


class DDRPrototype(AreaPrototype):
    """Class for a simple DDR area representation."""

    capacity_min = 2
    capacity_max = get_capacity_max(packing.VisuDataDirectDoubleRow, capacity_min)
    typename = "DDR"
    visu_class = packing.VisuDataDirectDoubleRow

    def __init__(self, capacity):
        super(DDRPrototype, self).__init__(capacity)


class LinePrototype(AreaPrototype):
    """Class for a simple Line area representation."""

    capacity_min = 2
    capacity_max = get_capacity_max(packing.VisuDataLine, capacity_min)

    typename = "L"
    visu_class = packing.VisuDataLine

    def __init__(self, capacity):
        super(LinePrototype, self).__init__(capacity)


OTHER_AREA_TYPES = {
    DSRPrototype: (DSR_90Prototype, DDRPrototype, LinePrototype),
    DSR_90Prototype: (DSRPrototype, DDRPrototype, LinePrototype),
    DDRPrototype: (DSRPrototype, DSR_90Prototype, LinePrototype),
    LinePrototype: (DSRPrototype, DSR_90Prototype, DDRPrototype),
}


# Do some precomputing

# Estimate the total maximum capacity (max with all areas of the same type
# with maximum capacity)
count_max_with_capacity_max_dsr, dims_dsr_cap_max = get_count_max_with_capacity_max(
    packing.VisuDataDirectSingleRow, DSRPrototype.capacity_min
)
(
    count_max_with_capacity_max_dsr_90,
    dims_dsr_90_cap_max,
) = get_count_max_with_capacity_max(
    packing.VisuDataDirectSingleRow_90, DSR_90Prototype.capacity_min
)
count_max_with_capacity_max_ddr, dims_ddr_cap_max = get_count_max_with_capacity_max(
    packing.VisuDataDirectDoubleRow, DDRPrototype.capacity_min
)
count_max_with_capacity_max_line, dims_line_cap_max = get_count_max_with_capacity_max(
    packing.VisuDataLine, LinePrototype.capacity_min
)
# count_max_single = max(count_max_dsr, count_max_ddr, count_max_line)
CAPACITY_MAX = max(
    dims_dsr_cap_max.count_inner,
    dims_dsr_90_cap_max.count_inner,
    dims_ddr_cap_max.count_inner,
    dims_line_cap_max.count_inner,
)

# Estimate the total maximum number of areas (max of all areas of the same
# type with minimum capacity)
count_max_with_capacity_min_dsr, dims_dsr_count_max = get_count_max_with_capacity_min(
    packing.VisuDataDirectSingleRow, DSRPrototype.capacity_min
)
(
    count_max_with_capacity_min_dsr_90,
    dims_dsr_90_count_max,
) = get_count_max_with_capacity_min(
    packing.VisuDataDirectSingleRow_90, DSR_90Prototype.capacity_min
)
count_max_with_capacity_min_ddr, dims_ddr_count_max = get_count_max_with_capacity_min(
    packing.VisuDataDirectDoubleRow, DDRPrototype.capacity_min
)
count_max_with_capacity_min_line, dims_line_count_max = get_count_max_with_capacity_min(
    packing.VisuDataLine, LinePrototype.capacity_min
)
COUNT_MAX = max(
    count_max_with_capacity_min_dsr,
    count_max_with_capacity_min_dsr_90,
    count_max_with_capacity_min_ddr,
    count_max_with_capacity_min_line,
)
COUNT_MIN = 1


def print_area_precomps():
    print(
        "Depot a: %d m, b: %d m"
        % (OC["scenario"]["DEPOT_A"], OC["scenario"]["DEPOT_B"])
    )
    print(
        "capacity_max_dsr: %d, count_max_dsr: %d, total slots: %d"
        % (
            dims_dsr_cap_max.items[0].count_inner,
            count_max_with_capacity_max_dsr,
            dims_dsr_cap_max.count_inner,
        )
    )
    print(
        "capacity_max_dsr_90: %d, count_max_dsr_90: %d, total slots: %d"
        % (
            dims_dsr_90_cap_max.items[0].count_inner,
            count_max_with_capacity_max_dsr_90,
            dims_dsr_90_cap_max.count_inner,
        )
    )
    print(
        "capacity_max_ddr: %d, count_max_ddr: %d, total slots: %d"
        % (
            dims_ddr_cap_max.items[0].count_inner,
            count_max_with_capacity_max_ddr,
            dims_ddr_cap_max.count_inner,
        )
    )
    print(
        "capacity_max_line: %d, count_max_line: %d, total slots: %d"
        % (
            dims_line_cap_max.items[0].count_inner,
            count_max_with_capacity_max_line,
            dims_line_cap_max.count_inner,
        )
    )


def init_random_area():
    """Return a new area prototype of random type with random capacity within
    bounds.
    """
    area_type = random.choice(
        [DSRPrototype, DSR_90Prototype, DDRPrototype, LinePrototype]
    )
    capacity = random.randint(area_type.capacity_min, area_type.capacity_max)
    return area_type(capacity)


class DepotPrototype:
    """Very basic depot representation."""

    def __init__(self):
        self.areas = []
        self._visu = None  # take care: needs reset after mutation
        self._ID = None
        self._ID_dm = None
        self.results = {"simulated": False}

    @property
    def visu(self):
        if self._visu is None:
            self._visu = packing.BinWithDistances(
                OC["scenario"]["DEPOT_A"], OC["scenario"]["DEPOT_B"], False
            )
            for area in self.areas:
                self._visu.items.append(area.visu)
        return self._visu

    @visu.setter
    def visu(self, value):
        self._visu = value

    @property
    def capacity(self):
        """Total parking area capacity."""
        return sum(area.capacity for area in self.areas)

    @property
    def ID(self):
        """Automatic ID based on self.areas."""
        if self._ID is None:
            self.ID = self.generate_ID()
        return self._ID

    @ID.setter
    def ID(self, value):
        self._ID = value

    def generate_ID(self):
        if self.areas:
            self.sort_areas()

            c = {
                DSRPrototype.typename: Counter(),
                DSR_90Prototype.typename: Counter(),
                DDRPrototype.typename: Counter(),
                LinePrototype.typename: Counter(),
            }
            for area in self.areas:
                c[area.typename][area.capacity] += 1

            ID = ""
            for typename, subcounter in c.items():
                for capacity, n in subcounter.items():
                    ID += "_" + str(n) + "x" + str(capacity) + typename
            ID = ID[1:]
            return ID
        else:
            return None

    @property
    def ID_dm(self):
        """ID with DSR and DDR merged."""
        if self._ID_dm is None:
            self.ID_dm = self.generate_ID_dm()
        return self._ID_dm

    @ID_dm.setter
    def ID_dm(self, value):
        self._ID_dm = value

    def generate_ID_dm(self):
        if self.areas:
            self.sort_areas()

            transl = {
                DSRPrototype.typename: "d",
                DSR_90Prototype.typename: "d",
                DDRPrototype.typename: "d",
                LinePrototype.typename: "l",
            }
            c = {
                DSRPrototype.typename: Counter(),
                DSR_90Prototype.typename: Counter(),
                DDRPrototype.typename: Counter(),
                LinePrototype.typename: Counter(),
            }
            for area in self.areas:
                c[area.typename][area.capacity] += 1

            # merge
            c[DSRPrototype.typename] += c[DDRPrototype.typename]
            del c[DDRPrototype.typename]

            ID = ""
            for typename, subcounter in c.items():
                for capacity, n in subcounter.items():
                    ID += "_" + str(n) + "x" + str(capacity) + transl[typename]
            ID = ID[1:]
            return ID
        else:
            return None

    def sort_areas(self):
        self.areas.sort(key=attrgetter("capacity"), reverse=True)
        self.areas.sort(key=attrgetter("typename"))

    def reset_results(self):
        """Reset of evaluation results that is required after changes such as
        crossover and mutation.
        """
        self.visu = None
        self.results.clear()
        self.results["simulated"] = False

    def __eq__(self, other):
        """Return True if DepotPrototype objects *self* and *other* have the
        same areas.
        """
        return self.ID == other.ID


def clone_depot(ind):
    """Return a new DepotPrototype instance with cloned areas.
    As function and not method to be applicable with map.
    """
    new = creator.Individual()
    new.areas = [area.clone() for area in ind.areas]
    # Keep the visu reference, simulation results and fitness for now. Must be
    # reset if anything changes before packing and simulation.
    new.visu = ind.visu
    new.results = deepcopy(ind.results)
    new.fitness = deepcopy(ind.fitness)
    return new


def init_random_depot(dcls):
    """Return a new DepotPrototype instance with a random amount of areas
    within bounds.
    """
    depot = dcls()
    area_count = random.randint(COUNT_MIN, COUNT_MAX)
    for i in range(area_count):
        depot.areas.append(init_random_area())
    return depot


if __name__ == "__main__":

    def _instantiate_areas():
        """Instantiate all possible areas within slot capacity bounds."""
        aid = {}
        ail = []
        for area_type in [DSRPrototype, DSR_90Prototype, DDRPrototype, LinePrototype]:
            aid[area_type] = {}
            for capacity in range(area_type.capacity_min, area_type.capacity_max + 1):
                aid[area_type][capacity] = area_type(capacity)
                ail.append(area_type(capacity))
        return aid, ail



# 1st: Namespaced Package Compat., see https://packaging.python.org/en/latest/guides/packaging-namespace-packages/
__path__ = __import__("pkgutil").extend_path(__path__, __name__)


# 2nd: Importing of the eflips modules -> TODO Cleanups here to avoid all those imports


import eflips.depot.layout_opt
import eflips.depot.settings_config
from eflips.depot.configuration import DepotConfigurator
from eflips.depot.depot import (
    DepotWorkingData,
    BackgroundStore,
    Depot,
    DepotControl,
    BaseArea,
    DirectArea,
    LineArea,
    AreaGroup,
    ParkingAreaGroup,
    DefaultActivityPlan,
    SpecificActivityPlan,
)
from eflips.depot.evaluation import DepotEvaluation
from eflips.depot.filters import VehicleFilter
from eflips.depot.input_epex_power_price import InputReader, PowerFrame
from eflips.depot.processes import (
    ProcessStatus,
    EstimateValue,
    Serve,
    ChargeAbstract,
    Charge,
    ChargeSteps,
    ChargeEquationSteps,
    Standby,
    Repair,
    Maintain,
    Precondition,
)
from eflips.depot.resources import DepotResource, DepotChargingInterface, ResourceSwitch
from eflips.depot.simple_vehicle import (
    VehicleType,
    VehicleTypeGroup,
    SimpleVehicle,
    SimpleBattery,
)
from eflips.depot.simulation import (
    DepotHost,
    SimulationHost,
    Depotinput,
    BaseMultipleSimulationHost,
)
from eflips.depot.smart_charging import SmartCharging, ControlSmartCharging
from eflips.depot.standalone import VehicleGenerator, SimpleTrip, Timetable
from eflips.depot.validation import Validator



"""Utility functions."""
import numpy as np


def share_nonzero(arr):
    """Share of elements that are unequal to zero in *arr*."""
    print(type(arr), arr)
    return np.count_nonzero(arr) / len(arr)


class StoppingCriteria:
    """Stopping utilities for optimization loops of an evolutionary algorithm."""

    def __init__(self):
        self.bestfit_history = {}
        self._criteria = []

        self.g = 0
        self.pop = None

    def select(self, *args):
        """Select the stopping criteria. Pass at least one stopping criteria
        method name as str. The same method cannot be selected twice.
        """
        for crit in args:
            try:
                self._criteria.append(getattr(self, crit))
            except AttributeError:
                raise ValueError("StoppingCriteria has no criterion named %s" % crit)

    def check(self, *args, **kwargs):
        """Return True if any of the selected stopping criteria returns True.
        Pass parameters as kwargs based on stopping criteria requirements.
        """
        return any(crit(*args, **kwargs) for crit in self._criteria)

    @staticmethod
    def max_gen_reached(g, ngen, *args, **kwargs):
        """Return True if *g* is higher than or equal to *ngen*."""
        result = g >= ngen
        if result:
            print("Stopped because of reaching max generation.")
        return result

    @staticmethod
    def max_fit_reached(hof, maxfitness, *args, **kwargs):
        """Return True if *maxfitness* is reached.

        hof: [deap.tools.HallOfFame]
        maxfitness: [deap.creator.Fitness...]
        """
        if hof.items:
            result = hof[0].fitness >= maxfitness
            if result:
                print("Stopped because of reaching max fitness.")
            return result
        else:
            return False

    def no_improvement(self, g, hof, improvement_interval, *args, **kwargs):
        """Return True if there has been no fitness improvement for the last
        *improvement_interval* generations.
        """
        self.bestfit_history[g] = hof[0].fitness if hof.items else None

        g_lookup = g - improvement_interval
        if g_lookup > 0:
            if self.bestfit_history[g_lookup] is not None:
                result = self.bestfit_history[g_lookup] >= hof[0].fitness
                if result:
                    print(
                        "Stopped because of no fitness improvement in the last %d generations."
                        % improvement_interval
                    )
                return result
            else:
                return False
        else:
            return False

    @staticmethod
    def feasible_found(pop, feasible, *args, **kwargs):
        """Return True if there is a feasible individual in the current
        population.

        pop: [list] population
        feasible: [function] returning the feasibility of an individual
        """
        for ind in pop:
            if feasible(ind):
                return True
        return False

    @staticmethod
    def higher_than(value, limit, *args, **kwargs):
        """Return True if *value* is higher than *limit*."""
        result = value > limit
        if result:
            print("Stopped because value %s is higher than %s." % (value, limit))
        return result


def attrbased_set(seq, attr):
    """Determine unique and duplicate objects in *seq* based on the comparison
    of attribute *attr*.
    """
    uniques_attr = set()
    uniques = []
    duplicates = []
    for obj in seq:
        if getattr(obj, attr) not in uniques_attr:
            uniques_attr.add(getattr(obj, attr))
            uniques.append(obj)
        else:
            duplicates.append(obj)
    return uniques, duplicates



# -*- coding: utf-8 -*-
"""
Created on Thu Sep  7 09:31:54 2017

@author: e.lauth, p.mundt

Components for the configuration of a depot.

"""
import json
import os
from os.path import basename
from typing import Dict

import eflips
from eflips.depot.depot import Depot, LineArea, ParkingAreaGroup, SpecificActivityPlan
from eflips.depot.filters import VehicleFilter
from eflips.depot.resources import DepotResource, DepotChargingInterface, ResourceSwitch
from eflips.depot.processes import ChargeSteps
from eflips.evaluation import DataLogger
from eflips.helperFunctions import load_json, save_json


class DepotConfigurator:
    """Utilities to:

    - create a depot
    - add/remove resources, resource switches, processes, areas, groups and plans to/from a depot
    - load and save a json depot template before simulation start.

    Before env.run is called, complete() must be called.

    Configuration-related errors are not raised. Instead, the current action is
    cancelled and a prepared error message is returned to enable the sender to
    decide on further actions (e.g. show an info popup in GUI).

    To add/remove areas to/from groups, use AreaGroup or ParkingAreaGroup
    methods.

    Attributes:

    :param filename_loaded: [None or str] filename of the imported template. None
        until loading. Stays the same even if templatename changes.
    :param templatename: [str] filename of the imported template, excluding the path.
        May be changed manually. Used to create export filenames.
    :param templatename_display: [str] "pretty" version of templatename.
    :param multiplied_areas: Map for connecting area IDs added with 'amount' with
        their resulting areas.

    """

    basemsg_invalid = "Invalid depot configuration."

    def __init__(self, env):
        self.env = env
        self.depot = Depot(self.env, "New depot")

        self.filename_loaded = None
        self.templatename = "New depot"
        self.templatename_display = ""

        self.multiplied_areas = {}

        self.completed = False

    @property
    def isvalid(self):
        """Check if the current depot configuration is valid.
        Return (True, None) if valid, otherwise (False, errormsg).
        """
        # Check if there is at least one area
        if not self.depot.areas:
            errormsg = (
                self.basemsg_invalid + "Cannot start simulation with empty depot. At "
                "least one Area must be specified."
            )
            return False, errormsg

        # Check if there is at least one parking area group
        if not self.depot.parking_area_groups:
            errormsg = (
                self.basemsg_invalid
                + "At least one parking area group must be specified."
            )
            return False, errormsg

        # Check if there are empty groups (invalid)
        for group in self.depot.groups.values():
            if not group.stores:
                errormsg = (
                    self.basemsg_invalid
                    + 'Group "%s" must contain at least one Area.' % group.ID
                )
                return False, errormsg

        # Check if there is a default plan
        if self.depot.default_plan is None:
            errormsg = self.basemsg_invalid + "A DefaultActivityPlan must be specified."
            return False, errormsg

        # Check if there are empty plans (invalid)
        if not self.depot.default_plan:
            errormsg = (
                self.basemsg_invalid
                + "DefaultActivityPlan %s cannot be empty." % self.depot.default_plan.ID
            )
            return False, errormsg
        for seq in self.depot.specific_plans:
            if not seq:
                errormsg = (
                    self.basemsg_invalid
                    + "SpecificActivityPlan %s cannot be empty." % seq.ID
                )
                return False, errormsg

        # Check if resource_switches have a resource and breaks
        for sw in self.depot.resource_switches.values():
            if sw.resource is None:
                errormsg = "No resource set for ResourceSwitch %s." % sw.ID
                return False, errormsg
            if not sw.breaks:
                errormsg = "ResourceSwitch %s breaks cannot be empty." % sw.ID
                return False, errormsg

        # Check if Precondition uses higher power than any charging interface
        # can provide
        for process in self.depot.processes.values():
            if process["typename"] == "Precondition":
                power_precond = process["kwargs"]["power"]
                for resource in self.depot.resources.values():
                    if isinstance(resource, DepotChargingInterface):
                        if resource.max_power < power_precond:
                            errormsg = (
                                "Charging interface %s max_power is lower than Precondition power."
                                % resource.ID
                            )
                            return False, errormsg
                duration_precond = process["kwargs"]["dur"]
                if (
                    duration_precond
                    > eflips.globalConstants["depot"]["lead_time_match"]
                ):
                    errormsg = "A Precondition time is higher then the lead_time_match (settings)"
                    return False, errormsg

        # Check if ChargeSteps uses higher power than the charging interfaces
        # at areas can provide
        for process in self.depot.processes.values():
            if process["typename"] == "ChargeSteps":
                max_power_charge = max(i[1] for i in process["kwargs"]["steps"])
                for area in self.depot.areas.values():
                    if process["kwargs"]["ID"] in area.available_processes:
                        for chint in area.charging_interfaces:
                            if chint.max_power < max_power_charge:
                                errormsg = (
                                    "Charging interface %s max_power is lower than ChargeSteps %s max power."
                                    % (chint.ID, process["kwargs"]["ID"])
                                )
                                return False, errormsg

        return True, None

    def reset(self):
        """Reset all depot configuration variables.
        Guarantees stable references for dicts and lists, e.g. a dict is
        cleared instead of being overwritten with an empty one.

        [untested, perhaps incomplete]
        """
        self.filename_loaded = None
        self.templatename = "New depot"
        self.templatename_display = ""

        self.multiplied_areas.clear()

        self.depot.ID = "New depot"
        self.depot.resources.clear()
        self.depot.resource_switches.clear()
        self.depot.processes.clear()
        self.depot.areas.clear()
        self.depot.groups.clear()
        self.depot.default_plan = None
        self.depot.specific_plans.clear()

        self.depot.direct_departure_areas.clear()
        self.depot.parking_area_groups.clear()

        self.depot.capacity = 0
        self.depot.parking_capacity = 0
        self.depot.parking_capacity_direct = 0

        self.depot.depot_control.departure_areas.clear()
        self.depot.depot_control.dispatch_strategy_name = "FIRST"

    def add_resource(self, typename, **kwargs):
        """Instantiate a DepotResource object and add it to self.depot. All
        required parameters have to be passed as keyword arguments (see class
        definition for documentation).
        Return (resource object, None) if successful, otherwise (None,
        errormsg).
        """
        ID = kwargs["ID"]
        # Check if ID is unique
        if ID in self.depot.resources:
            errormsg = 'Invalid ID "%s". IDs must be unique among resources.' % ID
            return None, errormsg

        # Instantiate DepotResource
        cls = getattr(eflips.depot.resources, typename)
        resource = cls(env=self.env, depot=self.depot, **kwargs)
        self.depot.resources[ID] = resource
        return resource, None

    def remove_resource(self, ID):
        """Remove resource with *ID*.
        Return removed_from_special [list], containing processes that the
        resource was removed from.
        """
        removed_from_special = []
        resource = self.depot.resources[ID]

        # From processes
        for process in self.depot.processes.values():
            if resource in process.required_resources:
                process.required_resources.remove(resource)
                removed_from_special.append(process)

        # From resource switches
        for switch in self.depot.resource_switches.values():
            if resource is switch.resource:
                switch.resource = None

        del self.depot.resources[ID]
        return removed_from_special

    @staticmethod
    def export_resource(resource):
        """Return a dict that represents the configuration of *resource*."""
        data = {"typename": type(resource).__name__}
        if isinstance(resource, DepotResource):
            data["capacity"] = resource.capacity
        if isinstance(resource, DepotChargingInterface):
            data["max_power"] = resource.max_power
        return data

    def add_resource_switch(self, **kwargs):
        """Instantiate a ResourceSwitch object and add it to self.depot. All
        required parameters have to be passed as keyword arguments (see class
        definition for documentation).
        Parameter "resource" must be a string. The reference is resolved here.
        Return (resource_switch object, None) if successful, otherwise (None,
        errormsg).
        """
        ID = kwargs["ID"]
        # Check if ID is unique
        if ID in self.depot.resource_switches:
            errormsg = (
                'Invalid ID "%s". IDs must be unique among ' "resource_switches." % ID
            )
            return None, errormsg

        # Check if resource exists
        if kwargs["resource"] not in self.depot.resources:
            errormsg = (
                'Resource "%s" not found, but used for Resource '
                'Switch "%s".' % (kwargs["resource"], ID)
            )
            return None, errormsg

        kwargs["resource"] = self.depot.resources[kwargs["resource"]]

        # Instantiate ResourceSwitch
        resource_switch = ResourceSwitch(env=self.env, **kwargs)
        self.depot.resource_switches[ID] = resource_switch

        return resource_switch, None

    def remove_resource_switch(self, ID):
        """Remove resource_switch with *ID*. A related resource is not deleted."""
        del self.depot.resource_switches[ID]
        return []

    @staticmethod
    def export_resource_switch(resource_switch):
        """Return a dict that represents the configuration of *resource_switch*."""
        return {
            "resource": resource_switch.resource.ID,
            "breaks": resource_switch.breaks,
            "preempt": resource_switch.preempt,
            "strength": resource_switch.strength_input,
        }

    def add_process(self, typename, **kwargs):
        """Prepare process data for instantiation, which happens during
        simulation.
        *typename* must be a key in eflips.depot.processes.
        Return (process data dict, None) if successful, otherwise (None,
        errormsg).
        """
        ID = kwargs["ID"]
        # Check if ID is unique
        if ID in self.depot.processes:
            errormsg = 'Invalid ID "%s". IDs must be unique among processes.' % ID
            return None, errormsg

        self.depot.processes[ID] = {}
        self.depot.processes[ID]["kwargs"] = kwargs

        # Convert typename from str to an actual class reference. The name is
        # preserved
        cls = getattr(eflips.depot.processes, typename)
        self.depot.processes[ID]["type"] = cls
        self.depot.processes[ID]["typename"] = typename

        if cls is ChargeSteps:
            ChargeSteps.check_steps(kwargs["steps"])

        # Instantiate VehicleFilter, if any
        if "vehicle_filter" in kwargs and kwargs["vehicle_filter"] is not None:
            vf = kwargs["vehicle_filter"]
            kwargs["vehicle_filter"] = VehicleFilter(env=self.env, **vf)

        # Convert required_resources from list of str to list of references
        if "required_resources" in kwargs and kwargs["required_resources"]:
            reqRes = []
            for resID in kwargs["required_resources"]:
                if resID in self.depot.resources:
                    reqRes.append(self.depot.resources[resID])
                else:
                    errormsg = (
                        'Resource "%s" not found, but used for '
                        'Process "%s".' % (resID, ID)
                    )
                    return None, errormsg
            kwargs["required_resources"] = reqRes

        return self.depot.processes[ID], None

    def remove_process(self, ID):
        """Remove process with *ID*. A related resource is not deleted."""
        removed_from_special = []
        # From areas
        for area in self.depot.areas.values():
            if ID in area.available_processes:
                area.available_processes.remove(ID)
                removed_from_special.append(area)

        del self.depot.processes[ID]
        return removed_from_special

    def export_process(self, procdata):
        """Return procdata in an export format."""
        # Create new dict where entries of subdict 'kwargs' are on first level
        procdata_export = {}
        procdata_export["typename"] = procdata["typename"]
        procdata_export.update(procdata["kwargs"])
        # Convert VehicleFilter object to dict representation
        procdata_export["vehicle_filter"] = self.export_vehicle_filter(
            procdata["kwargs"]["vehicle_filter"]
        )
        # Convert required_resources from list of references to list of str
        if "required_resources" in procdata["kwargs"]:
            procdata_export["required_resources"] = [
                res.ID for res in procdata["kwargs"]["required_resources"]
            ]
        del procdata_export["ID"]

        return procdata_export

    def add_area(self, typename, **kwargs):
        """Call self._add_area once if kwargs['amount'] is 1 or missing.
        If amount > 1, create sub-IDs and call self._add_area amount number of
        times. amount is removed from kwargs.
        Return (added_areas, None) if successful, otherwise (None, errormsg).
        added_areas is a list that contains all added areas from this method
        call.
        """
        origID = kwargs["ID"]
        # Check if ID is unique
        if (
            origID in self.depot.areas
            or origID in self.depot.groups
            or origID in self.multiplied_areas
        ):
            errormsg = (
                'Invalid area ID "%s". IDs must be unique among '
                "areas and groups." % origID
            )
            return None, errormsg

        # Check how many areas should be added (1 if amount is not given)
        amount = kwargs.pop("amount", 1)

        if amount > 1:
            raise ValueError(
                "Instantiation of areas with amount > 1 is disabled due to "
                "the direct assignment of charging interfaces."
            )
            # Add amount number of areas
            self.multiplied_areas[origID] = []

            for i in range(amount):
                ID_i = origID + "_" + str(i + 1)

                # Check if ID_i is unique
                if ID_i in self.depot.areas or ID_i in self.depot.groups:
                    errormsg = (
                        'Invalid area ID "%s". IDs must be '
                        "unique among areas and groups." % origID
                    )
                    return None, errormsg

                kwargs["ID"] = ID_i
                area, errormsg = self._add_area(typename, **kwargs)
                if area is None:
                    return None, errormsg
                else:
                    self.multiplied_areas[origID].append(area)

            return self.multiplied_areas[origID], None

        else:
            # Add one area
            area, errormsg = area, errormsg = self._add_area(typename, **kwargs)
            if area is None:
                return None, errormsg
            else:
                return [area], errormsg

    def _add_area(self, typename, **kwargs):
        """Instantiate a DirectArea or LineArea object and add it to
        self.depot. All required parameters have to be passed as keyword
        arguments (see class definition for documentation).
        Parameter 'available_processes' must be a list of str.
        Return (area, None) if successful, otherwise (None, errormsg).

        Should only be called through self.add_area.
        """
        ID = kwargs["ID"]

        # Instantiate VehicleFilter, if any
        vf = kwargs["entry_filter"]
        if vf is not None:
            kwargs["entry_filter"] = VehicleFilter(env=self.env, **vf)

        # Check for process validity. Entries stay str
        available_processes = kwargs["available_processes"]
        for procID in available_processes:
            if procID not in self.depot.processes:
                errormsg = 'Process "%s" not found, but used for Area ' '"%s".' % (
                    procID,
                    ID,
                )
                return None, errormsg

        # Add charging interfaces, if any
        if "charging_interfaces" in kwargs and kwargs["charging_interfaces"]:
            charging_interfaces = []
            for ciID in kwargs["charging_interfaces"]:
                if ciID in self.depot.resources:
                    charging_interfaces.append(self.depot.resources[ciID])
                else:
                    errormsg = 'Resource "%s" not found, but used for ' 'Area "%s".' % (
                        ciID,
                        ID,
                    )
                    return None, errormsg
            kwargs["charging_interfaces"] = charging_interfaces

        # Instantiate area
        cls = getattr(eflips.depot.depot, typename)
        area = cls(env=self.env, **kwargs)
        self.depot.areas[ID] = area
        area.depot = self.depot

        # Add area to departure_areas if issink
        if area.issink:
            self.depot.depot_control.departure_areas.add_store(area)

        return area, None

    def remove_area(self, ID):
        """Remove area with *ID*.
        Return removed_from_special [list], containing groups and plans
        the area was removed from (excluding departure_areas and
        multiplied_areas because they are background activities).
        Related processes are not deleted.
        """
        removed_from_special = []
        area = self.depot.areas[ID]

        # From departure_areas (included if issink)
        if area in self.depot.depot_control.departure_areas.stores:
            self.depot.depot_control.departure_areas.remove_store(area)

        # From groups including parking area groups
        for group in self.depot.groups.values():
            if area in group.stores:
                group.remove_store(area)
                removed_from_special.append(group)

        # From plans
        if self.depot.default_plan is not None and area in self.depot.default_plan:
            self.depot.default_plan.remove(area)
            removed_from_special.append(self.depot.default_plan)
        for seq in self.depot.specific_plans:
            if area in seq:
                seq.remove(area)
                removed_from_special.append(seq)

        # From self.multiplied_areas
        for areas in self.multiplied_areas.values():
            if area in areas:
                areas.remove(area)

        del self.depot.areas[ID]
        return removed_from_special

    def export_area(self, area):
        """Return a dict that represents the configuration of *area*."""
        data = {
            "typename": type(area).__name__,
            "capacity": area.capacity,
            "charging_interfaces": [ci.ID for ci in area.charging_interfaces]
            if area.charging_interfaces is not None
            else [],
            "available_processes": area.available_processes,
            "issink": area.issink,
            "entry_filter": self.export_vehicle_filter(area.entry_filter),
        }
        if isinstance(area, LineArea):
            data["side_put_default"] = area.side_put_default
            data["side_get_default"] = area.side_get_default
        return data

    def add_group(self, typename, **kwargs):
        """Instantiate an AreaGroup object and add it to self.depot. All
        required parameters have to be passed as keyword arguments (see class
        definition for documentation).
        Parameter stores must be a list of str.
        Return (group object, None) if successful, otherwise (None, errormsg).
        """
        ID = kwargs["ID"]
        # Check if ID is unique
        if (
            ID in self.depot.areas
            or ID in self.depot.groups
            or ID in self.multiplied_areas
        ):
            errormsg = (
                'Invalid group ID "%s". IDs must be unique among '
                "areas and groups." % ID
            )
            return None, errormsg

        # Prepare stores, a list that may be composed of regular and
        # multiplied areas and handed over at instantiation of the group
        stores = []
        for areaID in kwargs["stores"]:
            if areaID in self.depot.areas:
                # Entry is a single area
                stores.append(self.depot.areas[areaID])

            elif areaID in self.multiplied_areas:
                # Area was multiplied. Add its resulting areas.
                stores.extend(self.multiplied_areas[areaID].copy())
            else:
                errormsg = 'Area "%s" not found, but used for Group ' '"%s".' % (
                    areaID,
                    ID,
                )
                return None, errormsg

        # Get class reference
        cls = getattr(eflips.depot.depot, typename)

        # Do checks on area(s) specific for parking area groups
        if cls is ParkingAreaGroup:
            for area in stores:
                check, errormsg = self.check_area_for_group(area, ID)
                if not check:
                    return None, errormsg

        kwargs["stores"] = stores
        # Instantiate group
        group = cls(env=self.env, **kwargs)
        self.depot.groups[ID] = group

        group.depot = self.depot

        # Specific actions for parking area groups
        if cls is ParkingAreaGroup:
            self.depot.parking_area_groups.append(group)
            for area in group.stores:
                area.parking_area_group = group

        return group, None

    @staticmethod
    def check_area_for_group(area, groupID):
        """Do checks on the validity of areas in parking area groups. Helper
        function for self.add_group.
        Return (True, None) if successful, otherwise (False, errormsg).
        """
        # Check if issink is True for the area (mandatory in ParkingAreaGroup)
        if not area.issink:
            errormsg = (
                "Area %s cannot be added to parking area group %s. "
                "Parking area groups may only contain areas where "
                "issink is True." % (area.ID, groupID)
            )
            return False, errormsg

        # Check if area already belongs to another parking area group
        # (conflict)
        if area.parking_area_group is not None:
            errormsg = (
                'Area "%s" cannot be added to assignmentgroup "%s" '
                "because it's still a member of parking area group "
                '"%s"' % (area.ID, groupID, area.parking_area_group.ID)
            )
            return False, errormsg

        return True, None

    def remove_group(self, ID):
        """Remove group with *ID*.
        Return removed_from_special [list], containing plans the group
        was removed from. Related areas are not deleted.
        """
        removed_from_special = []
        group = self.depot.groups[ID]

        # Remove and areas from group and release if type is ParkingAreaGroup
        group.clear()

        # From depot.parking_area_groups
        if isinstance(group, ParkingAreaGroup):
            self.depot.parking_area_groups.remove(group)

        # From plans
        if self.depot.default_plan is not None and group in self.depot.default_plan:
            self.depot.default_plan.remove(group)
            removed_from_special.append(self.depot.default_plan)
        for seq in self.depot.specific_plans:
            if group in seq:
                seq.remove(group)
                removed_from_special.append(seq)

        del self.depot.groups[ID]
        return removed_from_special

    @staticmethod
    def export_group(group):
        """Return a dict that represents the configuration of *group*."""
        data = {
            "typename": type(group).__name__,
            "stores": [store.ID for store in group.stores],
        }
        if isinstance(group, ParkingAreaGroup):
            data["parking_strategy_name"] = group.parking_strategy_name
        return data

    def add_plan(self, typename, **kwargs):
        """Instantiate an AreaGroup (or subclass) object and add it to
        self.depot. All required parameters have to be passed as keyword
        arguments (see class definition for documentation).
        Return (plan object, None) if successful, otherwise (None, errormsg).
        """
        ID = kwargs["ID"]
        # Check if ID is unique
        if (
            self.depot.default_plan is not None
            and self.depot.default_plan.ID == ID
            or ID in self.depot.specific_plans
        ):
            errormsg = 'Invalid ID "%s". IDs must be unique among plans.' % ID
            return None, errormsg

        else:
            # Get list of areas and groups
            locations, errormsg = self.get_locations(kwargs["locations"])
            if locations is None:
                return None, errormsg
            kwargs["locations"] = locations

            # Instantiate VehicleFilter for specific plans
            if "vehicle_filter" in kwargs and kwargs["vehicle_filter"] is not None:
                if typename == "DefaultActivityPlan":
                    errormsg = (
                        "ActivityPlans of type Default cannot have " "a vehicle_filter."
                    )
                    return None, errormsg
                else:
                    vf = kwargs["vehicle_filter"]
                    kwargs["vehicle_filter"] = VehicleFilter(env=self.env, **vf)

            # Instantiate plan
            cls = getattr(eflips.depot.depot, typename)
            if typename == "DefaultActivityPlan":
                del kwargs["ID"]
                plan = cls(**kwargs)
                self.depot.default_plan = plan
            else:
                plan = cls(**kwargs)
                self.depot.specific_plans[ID] = plan
            return plan, None

    def get_locations(self, IDs):
        """Compose a list of areas and groups based on IDs. Helper function for
        self.add_plan.
        """
        locations = []
        for ID in IDs:
            if ID in self.depot.areas:
                locations.append(self.depot.areas[ID])
            elif ID in self.depot.groups:
                locations.append(self.depot.groups[ID])
            else:
                errormsg = 'Area or Group "%s" not found. ' % ID
                return None, errormsg
        return locations, None

    def remove_plan(self, ID):
        """Remove plan with *ID* from default or specific plans.
        Related areas and groups are not deleted.
        """
        if self.depot.default_plan is not None and self.depot.default_plan.ID == ID:
            self.depot.default_plan = None

        elif ID in self.depot.specific_plans:
            del self.depot.specific_plans[ID]

        return []

    def export_plan(self, plan):
        """Return a dict that represents the configuration of *plan*."""
        data = {
            "typename": type(plan).__name__,
            "locations": [element.ID for element in plan],
        }
        if isinstance(plan, SpecificActivityPlan):
            data["vehicle_filter"] = self.export_vehicle_filter(plan.vehicle_filter)
        return data

    @staticmethod
    def export_vehicle_filter(vf):
        """Return a dict that represents the configuration of *vf*. Return None
        if all vehicles are permitted.
        """
        if vf is None or not vf.filter_names:
            return None
        else:
            data = vars(vf)
            del data["filters"]
            if "vehicle_types" in data:
                data["vehicle_types"] = data["vehicle_types_str"]
                del data["vehicle_types_str"]
            return data

    def load(self, template: Dict | str):
        """
        Load a depot configuration from a template the template may either be a
        - dict containing the configuration
        - a string containing a path without extension to a json file

        :param template: dict or str
        """
        self.reset()

        if isinstance(template, str):
            loaded_data = load_json(template)
            self.filename_loaded = template
            self.templatename = template.split("\\")[-1]
        elif isinstance(template, dict):
            loaded_data = template
            self.filename_loaded = "No filename"
            self.templatename = "Not template name"
        else:
            raise TypeError(
                "template must be a path-like object defining a JSON file or a dict."
            )

        self.templatename_display = loaded_data["templatename_display"]
        self.depot.ID = loaded_data["general"]["depotID"]
        self.depot.depot_control.dispatch_strategy_name = loaded_data["general"][
            "dispatch_strategy_name"
        ]

        # Import resources
        for k in loaded_data["resources"]:
            data = loaded_data["resources"][k]
            resource, errormsg = self.add_resource(data.pop("typename"), ID=k, **data)
            if resource is None:
                return False, errormsg

        # Import resource_switches
        for k in loaded_data["resource_switches"]:
            data = loaded_data["resource_switches"][k]
            resource_switch, errormsg = self.add_resource_switch(ID=k, **data)
            if resource_switch is None:
                return False, errormsg

        # Import processes
        for k in loaded_data["processes"]:
            data = loaded_data["processes"][k]
            process, errormsg = self.add_process(data.pop("typename"), ID=k, **data)
            if process is None:
                return False, errormsg

        # Import areas
        for k in loaded_data["areas"]:
            data = loaded_data["areas"][k]
            area, errormsg = self.add_area(data.pop("typename"), ID=k, **data)
            if area is None:
                return False, errormsg

        # Import groups
        for k in loaded_data["groups"]:
            data = loaded_data["groups"][k]
            group, errormsg = self.add_group(data.pop("typename"), ID=k, **data)
            if group is None:
                return False, errormsg

        # Import activity plans
        for k in loaded_data["plans"]:
            data = loaded_data["plans"][k]
            plan, errormsg = self.add_plan(data.pop("typename"), ID=k, **data)
            if plan is None:
                return False, errormsg

        return True, None

    def save(self, filename):
        """Save current configuration as a json template. The configuration
        must be valid.
        *filename* must be suitable for eflips.settings.save_json.
        Return (True, None) if successful, otherwise (False, errormsg).
        """
        success, errormsg = self.isvalid
        if not success:
            return success, errormsg

        configuration = dict()
        configuration["templatename_display"] = self.templatename
        configuration["general"] = {
            "depotID": self.depot.ID,
            "dispatch_strategy_name": self.depot.depot_control.dispatch_strategy_name,
        }

        # Export resources
        configuration["resources"] = {
            k: self.export_resource(v) for k, v in self.depot.resources.items()
        }
        # Export resource_switches
        configuration["resource_switches"] = {
            k: self.export_resource_switch(v)
            for k, v in self.depot.resource_switches.items()
        }
        # Export processes
        configuration["processes"] = {
            k: self.export_process(v) for k, v in self.depot.processes.items()
        }
        # Export areas
        configuration["areas"] = {
            k: self.export_area(v) for k, v in self.depot.areas.items()
        }
        # Export groups
        configuration["groups"] = {
            k: self.export_group(v) for k, v in self.depot.groups.items()
        }
        # Export specific plans
        configuration["plans"] = {
            k: self.export_plan(v) for k, v in self.depot.specific_plans.items()
        }
        # Export default plan
        configuration["plans"][self.depot.default_plan.ID] = self.export_plan(
            self.depot.default_plan
        )

        save_json(configuration, filename)
        return True, None

    def complete(self):
        """Actions that must take place before the simulation starts, but may
        not be possible upon initial creation of the depot since the
        possibility to create an empty depot is required.
        Return (True, None) if successful, otherwise (False, errormsg).
        May be called only once before simulation start.
        """
        if self.completed:
            errormsg = "Method DepotConfigurator.complete can be called " "only once."
            return False, errormsg

        # Final check for validity
        success, errormsg = self.isvalid
        if not success:
            return success, errormsg

        self.depot.depot_control._complete()

        if eflips.globalConstants["general"]["LOG_ATTRIBUTES"]:
            self.depot.init_store.logger = DataLogger(
                self.env, self.depot.init_store, "BACKGROUNDSTORE"
            )

        # Create a list of direct areas that are in a parking area group
        # (performance tweak)
        for group in self.depot.parking_area_groups:
            self.depot.direct_departure_areas.extend(group.direct_areas)

        # Run resource_switches
        for sw in self.depot.resource_switches.values():
            if sw.breaks:
                self.env.process(sw.run_break_cycle())

        self.depot.capacity = sum(area.capacity for area in self.depot.list_areas)
        self.depot.parking_capacity = sum(
            pag.capacity for pag in self.depot.parking_area_groups
        )
        self.depot.parking_capacity_direct = sum(
            pag.capacity_direct for pag in self.depot.parking_area_groups
        )

        self.depot.any_process_cancellable_for_dispatch = any(
            procdata["kwargs"]["cancellable_for_dispatch"]
            for procdata in self.depot.processes.values()
        )

        self.completed = True
        return True, None
